/////////////////////////////////////////////////////////////////////////////
// Implementation for a MFC class encapsulation of the synchronous sockets

/*
 * Copyright (c) 2003 - 2005 by Daniel Weber (daniel@astolat.de)
 * All rights reserved.
 *
 * This is a SSL/TLS-implementation for Windows-Sockets in C++ written by
 * Daniel Weber (daniel@astolat.de) based an examples provided by
 * Martin Ziacek (martin.ziacek@blueyonder.co.uk).
 *
 * You are allowed to use, modify and redistribute this source-code for
 * commercial and non-commercial software as long as the following
 * conditions are met:
 * - Copyright remains bei Daniel Weber and redistributions of the source-
 *   code must retain this copyright-notice, list of conditions and
 *   disclaimer.
 * - Redistributions in binary form must mention Daniel Weber as author
 *   of parts of the SSL/TLS-implementations at program startup and/or
 *   in the documentation.
 *
 * This source-code is provided by Daniel Weber "as is" and any express or
 * implied warranties, including but not limited to the implied warranties
 * of merchantability and fitness for a particular purpose are disclaimed.
 * The authors or contributors are not liable for any damages arising out
 * of the use of this source-code, even if advised of the possibility of
 * such damage.
 */

#include "stdafx.h"

#include "SyncSocketEx.h"
#include "SocketException.h"

IMPLEMENT_DYNAMIC(CSyncSocketEx, CObject)

/////////////////////////////////////////////////////////////////////////////
// CSyncSocketEx Construction

CSyncSocketEx::CSyncSocketEx()
{
	m_hSocket = INVALID_SOCKET;
	m_nTimeout = 120;
}

CSyncSocketEx::CSyncSocketEx( const CSyncSocketEx& rSrc )
{
	m_hSocket = rSrc.m_hSocket;
	m_nTimeout = rSrc.m_nTimeout;
}

void CSyncSocketEx::operator=( const CSyncSocketEx& rSrc )
{
	m_hSocket = rSrc.m_hSocket;
	m_nTimeout = rSrc.m_nTimeout;
}

CSyncSocketEx::~CSyncSocketEx()
{
	Close();
}

BOOL CSyncSocketEx::Create(UINT nSocketPort, int nSocketType, LPCTSTR lpszSocketAddress)
{
	if( !Socket( nSocketType ) )
	{
		return FALSE;
	}

	if( !Bind( nSocketPort, lpszSocketAddress ) )
	{
		return FALSE;
	}

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CSyncSocketEx Attributes

CSyncSocketEx::operator SOCKET() const
{
	ASSERT( m_hSocket != INVALID_SOCKET );

	return m_hSocket;
}

BOOL CSyncSocketEx::Attach( SOCKET hSocket )
{
	ASSERT( hSocket != INVALID_SOCKET );

	m_hSocket = hSocket;

	return TRUE;
}

SOCKET CSyncSocketEx::Detach()
{
	SOCKET hSocket = m_hSocket;

	m_hSocket = INVALID_SOCKET;

	return INVALID_SOCKET;
}

BOOL CSyncSocketEx::GetPeerName(CString& rPeerAddress, UINT& rPeerPort)
{
	USES_CONVERSION;

	SOCKADDR_IN sockAddr;
	::memset( &sockAddr, 0, sizeof( sockAddr ) );

	int nSockAddrLen = sizeof( sockAddr );
	if( !GetPeerName( static_cast<SOCKADDR*>( &sockAddr ), &nSockAddrLen ) )
	{
		return FALSE;
	}

	rPeerPort = ::ntohs( sockAddr.sin_port );
	rPeerAddress = A2T( ::inet_ntoa( sockAddr.sin_addr ) );

	return TRUE;
}

BOOL CSyncSocketEx::GetPeerName(SOCKADDR* lpSockAddr, int* lpSockAddrLen)
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::getpeername( m_hSocket, lpSockAddr, lpSockAddrLen ) == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}

	return TRUE;
}

BOOL CSyncSocketEx::GetSockName(CString& rSocketAddress, UINT& rSocketPort)
{
	USES_CONVERSION;

	SOCKADDR_IN sockAddr;
	::memset( &sockAddr, 0, sizeof( sockAddr ) );

	int nSockAddrLen = sizeof( sockAddr );
	if( !GetSockName( static_cast<SOCKADDR*>( &sockAddr ), &nSockAddrLen ) )
	{
		return FALSE;
	}

	rSocketPort = ::ntohs( sockAddr.sin_port );
	rSocketAddress = A2T( ::inet_ntoa( sockAddr.sin_addr ) );

	return TRUE;
}

BOOL CSyncSocketEx::GetSockName(SOCKADDR* lpSockAddr, int* lpSockAddrLen)
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::getsockname( m_hSocket, lpSockAddr, lpSockAddrLen ) == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}

	return TRUE;
}

BOOL CSyncSocketEx::SetSockOpt(int nOptionName, const void* lpOptionValue, int nOptionLen, int nLevel)
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::setsockopt( m_hSocket, nLevel, nOptionName, static_cast<LPCSTR>( lpOptionValue ), nOptionLen ) == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
		return FALSE;
	}

	return TRUE;
}

BOOL CSyncSocketEx::GetSockOpt(int nOptionName, void* lpOptionValue, int* lpOptionLen, int nLevel)
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::getsockopt( m_hSocket, nLevel, nOptionName, static_cast<LPSTR>( lpOptionValue ), lpOptionLen ) == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CAscynSocket Operations

BOOL CSyncSocketEx::Accept( CSyncSocketEx& rConnectedSocket, SOCKADDR* lpSockAddr, int* lpSockAddrLen )
{
	SOCKET hSocket = ::accept( m_hSocket, lpSockAddr, lpSockAddrLen );
	if( hSocket == INVALID_SOCKET )
	{
		throw new CSocketException( GetLastError() );
	}
	
	rConnectedSocket.m_hSocket = hSocket;

	return TRUE;
}

BOOL CSyncSocketEx::Bind( UINT nSocketPort, LPCTSTR lpszSocketAddress )
{
	SOCKADDR_IN sockAddr;
	::memset( &sockAddr, 0, sizeof( sockAddr ) );

	// For correct operation of the T2A macro, see MFC Tech Note 59
	USES_CONVERSION;

	LPSTR lpszAscii = T2A( const_cast<LPSTR>( lpszSocketAddress ) );
	sockAddr.sin_family = AF_INET;

	if( lpszSocketAddress == NULL || strlen( lpszAscii ) == 0  )
	{
		sockAddr.sin_addr.s_addr = ::htonl( INADDR_ANY );
	}
	else
	{
		DWORD lResult = ::inet_addr( lpszAscii );
		if( lResult == INADDR_NONE )
		{
			throw new CSocketException( WSAEINVAL );
		}
		sockAddr.sin_addr.s_addr = lResult;
	}

	sockAddr.sin_port = ::htons( static_cast<u_short>( nSocketPort ) );

	return Bind( static_cast<SOCKADDR*>( &sockAddr ), sizeof( sockAddr ) );
}

BOOL CSyncSocketEx::Bind(const SOCKADDR* lpSockAddr, int nSockAddrLen)
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::bind( m_hSocket, lpSockAddr, nSockAddrLen ) == SOCKET_ERROR )
	{
		// Note: Wird im Aufbau des Sockets verwendet, Exceptions noch nicht sinnvoll
//		throw new CSocketException( GetLastError() );
		return FALSE;
	}

	return TRUE;
}

BOOL CSyncSocketEx::Close()
{
	if( m_hSocket == INVALID_SOCKET )
	{
		return FALSE;
	}

	if( ::closesocket( m_hSocket ) == SOCKET_ERROR )
	{
		// Note: Wird aus dem Destruktor aufgerufen, und dort sind Exceptions ja nicht so gut
//		throw new CSocketException( WSAEINVAL );
		return FALSE;
	}

	m_hSocket = INVALID_SOCKET;

	return TRUE;
}

BOOL CSyncSocketEx::Connect( LPCTSTR lpszHostAddress, UINT nHostPort )
{
	// Must have a valid host
	ASSERT( lpszHostAddress );
	ASSERT( _tcslen( lpszHostAddress ) );

	// For correct operation of the T2A macro, see MFC Tech Note 59
	USES_CONVERSION;

	SOCKADDR_IN sockAddr;
	::memset( &sockAddr, 0, sizeof( sockAddr ) );

	LPSTR lpszAscii = T2A( const_cast<LPSTR>( lpszHostAddress ) );

	// Determine if the address is in dotted notation
	sockAddr.sin_family = AF_INET;
	sockAddr.sin_addr.s_addr = inet_addr( lpszAscii );
	sockAddr.sin_port = ::htons( static_cast<u_short>( nHostPort ) );

	// If the address is not dotted notation, then go to DNS 
	if( sockAddr.sin_addr.s_addr == INADDR_NONE )
	{
		LPHOSTENT lphost = ::gethostbyname( lpszAscii );
		if( lphost != NULL )
		{
			sockAddr.sin_addr.s_addr = ( reinterpret_cast<LPIN_ADDR>( lphost->h_addr ) )->s_addr;
		}
		else
		{
			throw new CSocketException( WSAHOST_NOT_FOUND );
		}
	}

	// TODO: prfen ob hier wirklich explizit in dieser Klasse aufgerufen werden soll
	return CSyncSocketEx::Connect( static_cast<SOCKADDR*>( &sockAddr ), sizeof( sockAddr ) );
}

BOOL CSyncSocketEx::Connect( const SOCKADDR* lpSockAddr, int nSockAddrLen )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::connect( m_hSocket, lpSockAddr, nSockAddrLen ) == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}

	return TRUE;
}

BOOL CSyncSocketEx::IOCtl( long lCommand, DWORD* lpArgument )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::ioctlsocket( m_hSocket, lCommand, lpArgument ) == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}

	return TRUE;
}

BOOL CSyncSocketEx::Listen( int nConnectionBacklog )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::listen( m_hSocket, nConnectionBacklog ) == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}

	return TRUE;
}

int CSyncSocketEx::Receive( void* lpBuf, int nBufLen, int nFlags )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );

	if( !IsReadible( m_nTimeout ) )
	{
		throw new CSocketException( WSAECONNABORTED );
	}
  
	int n = ::recv( m_hSocket, static_cast<LPSTR>( lpBuf ), nBufLen, nFlags );
	if( n == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}
	return n;
}

int CSyncSocketEx::ReceiveFrom( void* lpBuf, int nBufLen, CString& rSocketAddress, UINT& rSocketPort, int nFlags )
{
	SOCKADDR_IN sockAddr;
	::memset( &sockAddr, 0, sizeof( sockAddr ) );

	int nSockAddrLen = sizeof( sockAddr );
	int nResult = ReceiveFrom( lpBuf, nBufLen, static_cast<SOCKADDR*>( &sockAddr ), &nSockAddrLen, nFlags );
	if( nResult != SOCKET_ERROR )
	{
		rSocketPort = ::ntohs( sockAddr.sin_port );
		rSocketAddress = ::inet_ntoa( sockAddr.sin_addr );
	}
	return nResult;
}

int CSyncSocketEx::ReceiveFrom( void* lpBuf, int nBufLen, SOCKADDR* lpSockAddr, int* lpSockAddrLen, int nFlags )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	int n = ::recvfrom( m_hSocket, static_cast<LPSTR>( lpBuf ), nBufLen, nFlags, lpSockAddr, lpSockAddrLen );
	if( n == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}
	return n;
}

BOOL CSyncSocketEx::ShutDown(int nHow )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	if( ::shutdown( m_hSocket, nHow ) == 0 SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}

	return TRUE;
}

int CSyncSocketEx::Send( const void* lpBuf, int nBufLen, int nFlags )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );

	if( !IsWritible( m_nTimeout ) )
	{
		throw new CSocketException( WSAECONNABORTED );
	}
  
	int n = ::send( m_hSocket, static_cast<LPCSTR>( lpBuf ), nBufLen, nFlags );
	if( n == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}
	return n;
}

int CSyncSocketEx::SendTo( const void* lpBuf, int nBufLen, UINT nHostPort, LPCTSTR lpszHostAddress, int nFlags )
{
	// Must have a valid host
	ASSERT( lpszHostAddress );
	ASSERT( _tcslen( lpszHostAddress ) );

	// For correct operation of the T2A macro, see MFC Tech Note 59
	USES_CONVERSION;

	SOCKADDR_IN sockAddr;
	::memset( &sockAddr, 0, sizeof( sockAddr ) );

	LPSTR lpszAscii = T2A( const_cast<LPTSTR>( lpszHostAddress ) );

	// Determine if the address is in dotted notation
	sockAddr.sin_family = AF_INET;

	if( lpszAscii == NULL )
	{
		sockAddr.sin_addr.s_addr = ::htonl( INADDR_BROADCAST );
	}
	else
	{
		sockAddr.sin_addr.s_addr = ::inet_addr( lpszAscii );
		if( sockAddr.sin_addr.s_addr == INADDR_NONE )
		{
			LPHOSTENT lphost = ::gethostbyname( lpszAscii );
			if( lphost != NULL )
			{
				sockAddr.sin_addr.s_addr = ( reinterpret_cast<LPIN_ADDR>( lphost->h_addr ) )->s_addr;
			}
			else
			{
				throw new CSocketException( WSAEINVAL );
			}
		}
	}

	sockAddr.sin_port = ::htons( static_cast<u_short>( nHostPort ) );

	return SendTo(lpBuf, nBufLen, static_cast<SOCKADDR*>( &sockAddr ), sizeof(sockAddr), nFlags);
}

int CSyncSocketEx::SendTo( const void* lpBuf, int nBufLen, const SOCKADDR* lpSockAddr, int nSockAddrLen, int nFlags )
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	int n = ::sendto( m_hSocket, static_cast<LPCSTR>( lpBuf ), nBufLen, nFlags, lpSockAddr, nSockAddrLen );
	if( n == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}
	return n;
}

/////////////////////////////////////////////////////////////////////////////
// CSyncSocketEx Implementation

BOOL CSyncSocketEx::Socket(int nSocketType, int nProtocolType, int nAddressFormat)
{
	if( m_hSocket != INVALID_SOCKET )
	{
		throw new CSocketException( WSAEADDRINUSE );
	}

	m_hSocket = ::socket( nAddressFormat, nSocketType, nProtocolType );
	if( m_hSocket == INVALID_SOCKET )
	{
		throw new CSocketException( GetLastError() );
	}

	return TRUE;
}

#ifdef _DEBUG
void CSyncSocketEx::AssertValid() const
{
}

void CSyncSocketEx::Dump(CDumpContext& dc) const
{
}

#endif

BOOL CSyncSocketEx::IsReadible( int nTimeout ) const
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	fd_set fds;
	FD_ZERO( &fds );
	FD_SET( m_hSocket, &fds );
	timeval tvTimeout = { nTimeout, 0 };

	int n = ::select( 0, &fds, NULL, NULL, ( nTimeout != -1 ? &tvTimeout : NULL ) );
	if( n == 1 )
	{
		return TRUE;
	}
	else if( n == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}

	return FALSE;
}

BOOL CSyncSocketEx::IsWritible( int nTimeout ) const
{
	// must have been created first
	ASSERT( m_hSocket != INVALID_SOCKET );
  
	fd_set fds;
	FD_ZERO( &fds );
	FD_SET( m_hSocket, &fds );
	timeval tvTimeout = { nTimeout, 0 };

	int n = ::select( 0, NULL, &fds, NULL, ( nTimeout != -1 ? &tvTimeout : NULL ) );
	if( n == 1 )
	{
		return TRUE;
	}
	else if( n == SOCKET_ERROR )
	{
		throw new CSocketException( GetLastError() );
	}

	return FALSE;
}
